#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>

#include "netSock.h"
#include "resource.h"

const unsigned char version  = 2;    // NET library v2
const unsigned char revision = 0;
const unsigned char build    = 1;    // increase that with each version

static char *libraryName      = "netSock Driver";

HINSTANCE hInst;
long isServer;
HWND ipClient;
HWND ipServer;
long ipAddress;
long sock;
short PortNum;
char PadSendData[128];
char Pad1Data[128];
char Pad2Data[128];
char PadSendSize;
char Pad1Size;
char Pad2Size;
int PadCount;
fd_set rset;
fd_set wset;
TIMEVAL tm = { 0, 0 };

unsigned long CALLBACK PSEgetLibType() {
	return PSE_LT_NET;
}

char* CALLBACK PSEgetLibName() {
	return libraryName;
}

unsigned long CALLBACK PSEgetLibVersion() {
	return version<<16|revision<<8|build;
}

void SysMessage(char *fmt, ...) {
	va_list list;
	char tmp[512];

	va_start(list,fmt);
	vsprintf(tmp,fmt,list);
	va_end(list);
	MessageBox(0, tmp, "netSock Msg", 0);
}

long CALLBACK NETinit() {
	WSADATA wsaData;
	int retval;

	isServer = 0;
    if ((retval = WSAStartup(0x101,&wsaData)) != 0) { 
        SysMessage("WSAStartup failed with error %d\n", retval); 
        WSACleanup(); 
        return -1; 
    } 

	return 0;
}

__inline int SEND(void *pData, int Size, int Mode) {
	int bytes;
	int count=0;

	if (Mode & 0x1) { // NONBLOCKING
		int ret;

		FD_ZERO(&wset);
		FD_SET(sock, &wset);

		ret = select(sock, NULL, &wset, NULL, &tm);
		if (ret == -1) return -1;
		if (FD_ISSET(wset.fd_array[0], &wset)) {
			return send(sock, pData, Size, 0);
		} else {
			return 0;
		}
	} else { // BLOCKING
		while (Size > 0) {
			bytes = send(sock, pData, Size, 0);
			if (bytes < 0) return -1;
			(char*)pData+= bytes; Size-= bytes;
			count+= bytes;
		}
	}

	return count;
}

__inline int RECV(void *pData, int Size, int Mode) {
	int bytes;
	int count=0;

	if (Mode & 0x1) { // NONBLOCKING
		int ret;

		FD_ZERO(&rset);
		FD_SET(sock, &rset);

		ret = select(sock, &rset, NULL, NULL, &tm);
		if (FD_ISSET(rset.fd_array[0], &rset)) {
			return recv(sock, pData, Size, 0);
		} else {
			return 0;
		}
	} else { // BLOCKING
		while (Size > 0) {
			bytes = recv(sock, pData, Size, 0);
			(char*)pData+= bytes; Size-= bytes;
			count+= bytes;
		}
	}

	return count;
}

void CopyIpToClipboard(HWND hW) {
	char str[256];
	struct hostent *host;
	HGLOBAL pMem;
	char *ptr;
	unsigned char *addr;

	gethostname(str, 256);
	host = gethostbyname(str);
	addr = host->h_addr_list[0];
	sprintf(str, "%d.%d.%d.%d", addr[0], addr[1], addr[2], addr[3]);

	pMem = GlobalAlloc(0, strlen(str)+1);

	ptr = GlobalLock(pMem);
	memcpy(ptr, str, strlen(str)+1);
	GlobalUnlock(pMem);

	OpenClipboard(hW);
	SetClipboardData(CF_TEXT, pMem);
	CloseClipboard();
}

BOOL CALLBACK OpenDlgProc(HWND hW, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	char str[256];

	switch(uMsg) {
		case WM_INITDIALOG:
			ipClient = GetDlgItem(hW, IDC_IPCLIENT);
			ipServer = GetDlgItem(hW, IDC_IPSERVER);

			SendMessage(ipClient, IPM_SETADDRESS, 0, MAKEIPADDRESS(127, 0, 0, 1));
			SendMessage(ipServer, IPM_SETADDRESS, 0, MAKEIPADDRESS(127, 0, 0, 1));

			CheckDlgButton(hW, IDC_RADIOS, TRUE);
			CheckDlgButton(hW, IDC_RADIOC, FALSE);

			sprintf(str, "%d", 33006);
			Edit_SetText(GetDlgItem(hW, IDC_PORT), str);
			return TRUE;

		case WM_COMMAND:
			switch(LOWORD(wParam)) {
				case IDC_COPYIP:
					CopyIpToClipboard(hW);
					return TRUE;

				case IDC_RADIOS: 
					CheckDlgButton(hW, IDC_RADIOC, FALSE);
					return TRUE;

				case IDC_RADIOC: 
					CheckDlgButton(hW, IDC_RADIOS, FALSE);
					return TRUE;

				case IDOK:
					Edit_GetText(GetDlgItem(hW, IDC_PORT), str, 256);
					PortNum = (short)atol(str);

					if (Button_GetCheck(GetDlgItem(hW, IDC_RADIOC))) {
						SendMessage(ipClient, IPM_GETADDRESS, 0, (LPARAM)&ipAddress);
						isServer = 0;
					} else {
						SendMessage(ipServer, IPM_GETADDRESS, 0, (LPARAM)&ipAddress);
						isServer = 1;
					}
					EndDialog(hW, TRUE);
					return TRUE;

				case IDCANCEL:
					EndDialog(hW, FALSE);
					return TRUE;
			}
	}
	return FALSE;
}

void DlgUpdate() {
    MSG msg;

    while (PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
}

int PauseCancel;

BOOL CALLBACK PauseDlgProc(HWND hW, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch(uMsg) {
		case WM_INITDIALOG:
			PauseCancel = 0;
			return TRUE;

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
       			case IDCANCEL:
					PauseCancel = 1;
					return TRUE;
			}
	}

	return FALSE;
}

int ShowPauseDlg() {
	HWND hW;

	ShowCursor(TRUE);

	hW = CreateDialog(hInst, MAKEINTRESOURCE(IDD_PAUSEDLG), GetActiveWindow(), PauseDlgProc);
	ShowWindow(hW, SW_SHOW);

	for (;;) {
		int Code;

		if (PauseCancel == 1) break;
		DlgUpdate();

		if (RECV(&Code, 1, 1) == 1) break;
	}

	DestroyWindow(hW);

	return PauseCancel;
}

int WaitCancel;

BOOL CALLBACK WaitDlgProc(HWND hW, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch(uMsg) {
		case WM_INITDIALOG:
			WaitCancel = 0;
			return TRUE;

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
       			case IDCANCEL:
					WaitCancel = 1;
					return TRUE;
			}
	}

	return FALSE;
}

long CALLBACK NETopen(HWND hWnd) {
	HWND hW;
	int count=0;
	struct sockaddr_in address;
	int _sock=-1;
	INITCOMMONCONTROLSEX icc;
	char ip[32];

	icc.dwICC = ICC_INTERNET_CLASSES;
	icc.dwSize = sizeof(INITCOMMONCONTROLSEX);
	InitCommonControlsEx(&icc);

    if (DialogBox(hInst,
                  MAKEINTRESOURCE(IDD_OPEN),
                  hWnd,  
                  (DLGPROC)OpenDlgProc) == FALSE) return -1;

	sprintf(ip, "%d.%d.%d.%d", FIRST_IPADDRESS(ipAddress), SECOND_IPADDRESS(ipAddress), THIRD_IPADDRESS(ipAddress), FOURTH_IPADDRESS(ipAddress));

	if (isServer) {
		int listen_sock, reuse_addr = 1;
		int ret;

		memset((char *) &address, 0, sizeof (address));

		address.sin_family = AF_INET;
		address.sin_port = PortNum;
		address.sin_addr.s_addr = INADDR_ANY;

		listen_sock = socket (AF_INET,SOCK_STREAM,0);
		if (listen_sock == -1)
			return -1;

		setsockopt(listen_sock, SOL_SOCKET, SO_REUSEADDR, (const char*)&reuse_addr
					,sizeof (reuse_addr));

		if (bind(listen_sock,(struct sockaddr *) &address,sizeof (address)) == -1)
			return -1;

		if (listen(listen_sock, 1) != 0) 
			return -1; 

		sock = -1;

		hW = CreateDialog(hInst, MAKEINTRESOURCE(IDD_WAIT), hWnd, WaitDlgProc);
		ShowWindow(hW, SW_SHOW);

		while (sock < 0) {
			FD_ZERO(&rset);
			FD_SET(listen_sock, &rset);

			ret = select(sock, &rset, NULL, NULL, &tm);
			if (FD_ISSET(rset.fd_array[0], &rset)) {
				sock = accept(listen_sock, NULL, NULL);    
			}

			if (WaitCancel == 1) break;
			DlgUpdate();
		}

		DestroyWindow(hW);
		if (WaitCancel == 1) return -1;
	} else {
		memset((char *)&address, 0, sizeof(address));
		address.sin_family = AF_INET;
		address.sin_port = PortNum;
		address.sin_addr.s_addr = inet_addr(ip);

		sock = socket(AF_INET, SOCK_STREAM, 0);

		if (connect(sock, (struct sockaddr *)&address,sizeof(address))!=0) {
			SysMessage("error connecting to %s\n", ip);
			return -1;
		}
	}

	PadCount = 0;
	Pad1Size = -1;
	Pad2Size = -1;

	return 0;
}

long CALLBACK NETclose() {
	closesocket(sock);

	return 0;
}

long CALLBACK NETshutdown() {
	WSACleanup(); 
	return 0;
}

void CALLBACK NETpause() {
	unsigned char Code = 0x80;

	SEND(&Code, 1, 0);
}

void CALLBACK NETresume() {
	unsigned char Code = 0x80;

	SEND(&Code, 1, 0);
}

long CALLBACK NETsendData(void *pData, int Size, int Mode) {
	return SEND(pData, Size, Mode);
}

long CALLBACK NETrecvData(void *pData, int Size, int Mode) {
	return RECV(pData, Size, Mode);
}

long CALLBACK NETsendPadData(void *pData, int Size) {
//	printf("NETsendPadData %d, %x\n", Size, ((char*)pData)[4]);

	if (PadCount == 0) {
		memcpy(PadSendData, pData, Size);
		PadSendSize = Size;
	
		if (SEND(&PadSendSize, 1, 0) == -1)
			return -1;
		if (SEND(pData, PadSendSize, 0) == -1)
			return -1;
	}

	return 0;
}

long CALLBACK NETrecvPadData(void *pData, int Pad) {
//	printf("NETrecvPadData %d (server=%d), %x\n", Pad, isServer, Pad1Data[4]);

	if (PadCount == 0 ||
		Pad2Size == -1 ||
		Pad1Size == -1) {
		if ((isServer == 1 && Pad == 1) ||
			(isServer == 0 && Pad == 2)) {
			memcpy(Pad1Data, PadSendData, PadSendSize);
			Pad1Size = PadSendSize;
			memcpy(pData, Pad1Data, Pad1Size);
		} else {
			if (RECV(&Pad2Size, 1, 0) == -1)
				return -1;
			if (Pad2Size & 0x80) { // pause code
				if (ShowPauseDlg() == 1)
					return -1;
			}
			if (RECV(pData, Pad2Size, 0) == -1)
				return -1;
			memcpy(Pad2Data, pData, Pad2Size);
		}
	} else {
		if (Pad == 1) {
			memcpy(pData, Pad1Data, Pad1Size);
		} else {
			memcpy(pData, Pad2Data, Pad2Size);
		}
	}

	if (Pad == 2) {
		PadCount++;
		if (PadCount == 16) PadCount = 0;
	}

//	printf("Pad %d: %x\n", Pad, ((char*)pData)[4]);

	return 0;
}

long CALLBACK NETqueryServer() {
	return isServer;
}

void CALLBACK NETconfigure() {
	SysMessage("Nothing to Configure");
}

BOOL CALLBACK AboutDlgProc(HWND hW, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch(uMsg) {
		case WM_INITDIALOG:
			return TRUE;

		case WM_COMMAND:
			switch(LOWORD(wParam)) {
				case IDOK:
					EndDialog(hW, FALSE);
					return TRUE;
			}
	}
	return FALSE;
}

void CALLBACK NETabout() {
    DialogBox(hInst,
              MAKEINTRESOURCE(IDD_ABOUT),
              GetActiveWindow(),  
              (DLGPROC)AboutDlgProc);
}

long CALLBACK NETtest() {
	return 0;
}

BOOL APIENTRY DllMain(HANDLE hModule,                  // DLL INIT
                      DWORD  dwReason, 
                      LPVOID lpReserved) {
	hInst = (HINSTANCE)hModule;
	return TRUE;                                          // very quick :)
}
